
/*
 * (C) Copyright 2018 V-Key Solutions.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Contributors:
 *     Terence Peh
 *
 */

package vkey.android.cordova;

import android.app.Activity;
import android.content.BroadcastReceiver;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.os.Parcelable;
import android.util.Log;

import com.vkey.android.internal.vguard.engine.BasicThreatInfo;
import com.vkey.android.vguard.LocalBroadcastManager;
import com.vkey.android.vguard.VGuardBroadcastReceiver;
import com.vkey.android.vtap.VTapFactory;
import com.vkey.android.vtap.VTapInterface;
import com.vkey.android.vtap.pki.DistinguishedName;
import com.vkey.android.vtap.utility.ResultCode;

import org.apache.cordova.CallbackContext;
import org.apache.cordova.CordovaPlugin;
import org.apache.cordova.PluginResult;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static com.vkey.android.vtap.VTapInterface.PKI_FUNC_ID_AUTH;
import static com.vkey.android.vtap.VTapInterface.PKI_FUNC_ID_DOCSIGN;
import static com.vkey.android.vtap.VTapInterface.PKI_FUNC_ID_V_MESSAGE;

public class VTapSetupPlugin {

    private static volatile VTapSetupPlugin sInstance;

    private CordovaPlugin mPlugin;

    public static final String PUSH_NOTIFICATION_MESSAGE_ID_KEY         = "messageId";
    public static final String PUSH_NOTIFICATION_MESSAGE_TYPE_KEY       = "messageType";
    public static final String PUSH_NOTIFICATION_NOTIFY_MSG_FLAG_KEY    = "notifyMsgFlag";
    public static final String PUSH_NOTIFICATION_NOTIFY_MSG_KEY         = "notifyMsg";
    public static final String PUSH_NOTIFICATION_MSG_FLAG_KEY           = "msgFlag";
    public static final String PUSH_NOTIFICATION_PASS_TYPE_KEY          = "passType";

    public static final String PASS_TYPE_1                              = "1"; // Only Push Notification
    public static final String PASS_TYPE_2                              = "2"; // Push Notification + Message download

    public static final String NOTIFY_MSG_FLAG_CLEAR                    = "0";
    public static final String NOTIFY_MSG_FLAG_ENCRYPTED                = "1";

    public static final String MESSAGE_TYPE_ASP_CERT                    = "ASP_CERT";
    public static final String MESSAGE_TYPE_SMP_CERT                    = "SMP_CERT";
    public static final String MESSAGE_TYPE_ASP_DOC_CERT                = "ASP_DOC_CERT";
    public static final String MESSAGE_TYPE_AUTH                        = "AUTH";
    public static final String MESSAGE_TYPE_DOC_SIGN                    = "DOCSIGN";
    public static final String MESSAGE_TYPE_ASP_CERT_RENEW              = "ASP_CERT_RENEW";
    public static final String MESSAGE_TYPE_SMP_CERT_RENEW              = "SMP_CERT_RENEW";
    public static final String MESSAGE_TYPE_ASP_DOC_CERT_RENEW          = "ASPDOC_CERT_RENEW";
    public static final String MESSAGE_TYPE_TOKEN_REPROV                = "TOKEN_REPROV";
    public static final String MESSAGE_TYPE_SMP_MSG                     = "SMP_MSG";

    public static final String NOTIFY_MSG_BYTES                         = "notifyMsgBytes";
    public static final String DATA_TO_BE_SIGNED                        = "dataToBeSigned";
    public static final String TEXT_TO_DISPLAY                          = "textToDisplay";

    public static final String PUSH_NOTIFICATION_BROADCAST              = "push_notification";
    public static final String CERT_DOWNLOAD_RESULT                     = "certDownloadResult";
    public static final String REGISTRATION_TOKEN                       = "registrationToken";
    static final int DEVICE_CHECK                                       = 100;
    static final int PROVISIONING                                       = 101;
    static final int PROVISIONING_DATA                                  = 102;

    private static final String VTAP_SETUP_ACTION                       = "vkey.android.vtap.VTAP_SETUP";
    private static final String VTAP_SETUP_STATUS                       = "vkey.android.vtap.VTAP_SETUP_STATUS";
    private static final String PROFILE_LOADED_ACTION                   = "vkey.android.vguard.PROFILE_LOADED";
    private static final String VOS_READY_ACTION                        = "vkey.android.vguard.VOS_READY";
    private static final String VOS_FIRMWARE_RETURN_CODE_KEY            = "vkey.android.vguard.FIRMWARE_RETURN_CODE";
    public static final String PROFILE_THREAT_RESPONSE                  = "vkey.android.vguard.PROFILE_THREAT_RESPONSE";

    private static final String CB_SCAN_COMPLETE                        = "cb_scan_complete";
    private static final String CB_VTAP_SUCCESS                         = "cb_vtap_success";
    private static final String CB_VTAP_FAILED                          = "cb_vtap_failed";
    private static final String CB_VTAP_ERROR                           = "cb_vtap_error";

    private static final String CB_SET_HOSTNAME                         = "cb_set_hostname";
    private static final String CB_GET_TRUSTED_TIME                     = "cb_get_trusted_time";
    private static final String CB_DEVICE_COMPATIBILITY                 = "cb_device_compatibility";
    private static final String CB_SEND_DEVICE_INFO                     = "cb_send_device_info";
    private static final String CB_PROVISIONING_DONE                    = "cb_provisioning_done";
    private static final String CB_ALL_PROVISIONED_TOKEN                = "cb_all_provisioned_token";
    private static final String CB_TOKEN_TYPE                           = "cb_token_type";
    private static final String CB_DFP_HASH                             = "cb_dfp_hash";
    private static final String CB_GET_LOAD_ACK_TOKEN_FIRMWARE          = "cb_get_load_ack_token_firmware";
    private static final String CB_LOAD_TOKEN_FIRMWARE                  = "cb_load_token_firmware";
    private static final String CB_VALIDATE_CHECKSUM                    = "cb_validate_checksum_true";
    private static final String CB_GET_PKI_TOKEN_SERIAL                 = "get_pki_token_serial";
    private static final String CB_GET_TOKEN_SERIAL                     = "cb_get_token_serial";
    private static final String CB_TOKEN_FIRMWARE_VERSION               = "cb_token_firmware_version";
    private static final String CB_REMOVE_TOKEN_FIRMWARE                = "cb_remove_token";
    private static final String CB_LOAD_TOKEN                           = "cb_load_token";
    private static final String CB_UNLOAD_TOKEN                         = "cb_unload_token";
    private static final String CB_CREATE_TOKEN_PIN                     = "cb_set_token_pin";
    private static final String CB_TOKEN_PIN_REMEMBERED                 = "cb_token_pin_remembered";
    private static final String CB_CHECK_TOKEN_PIN                      = "cb_check_token_pin";
    private static final String CB_CHANGE_TOKEN_PIN                     = "cb_change_token_pin";
    private static final String CB_TOKEN_REGISTERED                     = "cb_token_registered";
    private static final String CB_SET_OTP_LENGTH                       = "cb_set_otp_length";
    private static final String CB_GENERATE_TOTP                        = "cb_generate_totp";
    private static final String CB_GENERATE_CR                          = "cb_generate_cr";
    private static final String CB_GENERATE_TXS                         = "cb_generate_txs";
    private static final String CB_GENERATE_CSR_AND_SEND                = "cb_generate_csr_and_send";
    private static final String CB_TRIGGER_CERT_PUSH_NOTIFICATION       = "cb_trigger_cert_push_notification";
    private static final String CB_PKI_CERT_DOWNLOAD                    = "cb_pki_cert_download";
    private static final String CB_GENERATE_CSR_AND_SEND_SYNC           = "cb_generate_csr_and_send_sync";
    private static final String CB_PKI_FUNCTION_REGISTERED              = "cb_pki_function_registered";
    private static final String CB_PKI_FUNCTION_PIN_REMEMBERED          = "cb_pki_function_pin_remembered";
    private static final String CB_PKI_FUNCTION_CHECK_PIN               = "cb_pki_function_check_pin";
    private static final String CB_REMOVE_PKI_FUNCTION                  = "cb_remove_pki_function";
    private static final String CB_PUSH_NOTIFICATION                    = "cb_push_notification";
    private static final String CB_REGISTERED_USER_ID                   = "cb_registered_user_id";
    private static final String CB_REGISTERED_DEVICE_ID                 = "cb_registered_device_id";
    private static final String CB_PKI_FUNCTION_AUTHENTICATE            = "cb_pki_function_authenticate";
    private static final String CB_VMESSAGE_DECRYPT                     = "cb_vmessage_decrypt";
    private static final String CB_VMESSAGE_DOWNLOAD                    = "cb_vmessage_download";
    private static final String CB_VMESSAGE_ACK                         = "cb_message_ack";
    private static final String CB_SET_ADDITIONAL_DATA                  = "cb_set_additional_data";
    private static final String CB_GET_ADDITIONAL_DATA                  = "cb_get_additional_data";
    private static final String CB_DISABLE_ADDITIONAL_DATA              = "cb_disable_additional_data";
    private static final String CB_SET_MESSAGE_DATA_TYPE                = "cb_set_message_data_type";
    private static final String CB_VTAP_PAUSE                           = "cb_vtap_pause";
    private static final String CB_VTAP_RESUME                          = "cb_vtap_resume";
    private static final String CB_VTAP_DESTROY                         = "cb_vtap_destroy";
    private static final String CB_GET_TROUBLESHOOTING_ID               = "cb_get_troubleshooting_id";
    private static final String CB_SEND_TROUBLESHOOTING_LOGS            = "cb_send_troubleshooting_logs";
    private static final String CB_CERT_DOWNLOAD                        = "cb_cert_download";

    private static final String RET_MESSAGE_SUCCESS                     = "successful";

    private VTapInterface iVTapManager;
    private ArrayList<String> provisioningInfo                          = new ArrayList<String>();

    private String mMessageId, mMessageType, mNotifyMsgFlag, mNotifyMsg, mMsgFlag, mPassType, mDataToBeSigned;
    private byte[] mNotifyMsgBytes;

    private CallbackContext mCallbackContext;
    //BroadCastReceiver fllag
    private boolean isRegistered;

    private static String TAG                                           = VTapSetupPlugin.class.getSimpleName();

    private VTapSetupPlugin(CordovaPlugin plugin) {
        mPlugin = plugin;
    }

    private BroadcastReceiver mVtapRcvr = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            if (VTAP_SETUP_ACTION.equalsIgnoreCase(intent.getAction())) {
                int vtapSetupStatus = intent.getIntExtra(VTAP_SETUP_STATUS, 0);

                if (vtapSetupStatus == ResultCode.VTAP_SETUP_SUCCESS) {
                    if(mCallbackContext != null) {
                        sendPluginResult(mCallbackContext, PluginResult.Status.OK, vtapSetupStatus);
                        Log.d(TAG, "adb_test vtap setup success..");
                    }
                } else {
                    if(mCallbackContext != null) {
                        sendPluginResult(mCallbackContext, PluginResult.Status.ERROR, vtapSetupStatus);
                        Log.e(TAG, "V-Tap initialization failed" + vtapSetupStatus);
                    }
                }
            }
        }
    };

    private BroadcastReceiver mPushRcvr = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            if (PUSH_NOTIFICATION_BROADCAST.equalsIgnoreCase(intent.getAction())) {
                mMessageId         = intent.getStringExtra(PUSH_NOTIFICATION_MESSAGE_ID_KEY);
                mMessageType       = intent.getStringExtra(PUSH_NOTIFICATION_MESSAGE_TYPE_KEY);
                mNotifyMsgFlag     = intent.getStringExtra(PUSH_NOTIFICATION_NOTIFY_MSG_FLAG_KEY);
                mNotifyMsg         = intent.getStringExtra(PUSH_NOTIFICATION_NOTIFY_MSG_KEY);
                mMsgFlag           = intent.getStringExtra(PUSH_NOTIFICATION_MSG_FLAG_KEY);
                mPassType          = intent.getStringExtra(PUSH_NOTIFICATION_PASS_TYPE_KEY);
                mNotifyMsgBytes    = intent.getByteArrayExtra(NOTIFY_MSG_BYTES);
                mDataToBeSigned    = intent.getStringExtra(DATA_TO_BE_SIGNED);

                Log.d(TAG, "mMessageId: " + mMessageId);
                Log.d(TAG, "mMessageType: " + mMessageType);
                Log.d(TAG, "mNotifyMsgFlag: " + mNotifyMsgFlag);
                Log.d(TAG, "mNotifyMsg: " + mNotifyMsg);
                Log.d(TAG, "mMsgFlag: " + mMsgFlag);
                Log.d(TAG, "mPassType: " + mPassType);
                Log.d(TAG, "mNotifyMsgBytes: " + mNotifyMsgBytes);
                Log.d(TAG, "mDataToBeSigned: " + mDataToBeSigned);

                if(VTapSetupPlugin.MESSAGE_TYPE_ASP_CERT.equalsIgnoreCase(mMessageType) ||
                        VTapSetupPlugin.MESSAGE_TYPE_SMP_CERT.equalsIgnoreCase(mMessageType) ||
                        VTapSetupPlugin.MESSAGE_TYPE_ASP_DOC_CERT.equalsIgnoreCase(mMessageType)) {

                    String sCertDownloadResult = intent.getStringExtra(CERT_DOWNLOAD_RESULT);
                    int certDownloadResult = Integer.parseInt(sCertDownloadResult);

                    if(certDownloadResult == ResultCode.VTAP_PKI_CERT_DOWNLOAD_SUCCESS) {
                        JSONObject obj = getSuccessCbValues(CB_CERT_DOWNLOAD, String.valueOf(certDownloadResult));
                        sendPluginResult(mCallbackContext, PluginResult.Status.OK, obj);
                    }
                    else {
                        if(VTapSetupPlugin.MESSAGE_TYPE_ASP_CERT.equalsIgnoreCase(mMessageType)) {
                            iVTapManager.removePKIFunction(PKI_FUNC_ID_AUTH);
                        }
                        if(VTapSetupPlugin.MESSAGE_TYPE_SMP_CERT.equalsIgnoreCase(mMessageType)) {
                            iVTapManager.removePKIFunction(PKI_FUNC_ID_V_MESSAGE);
                        }
                        else if(VTapSetupPlugin.MESSAGE_TYPE_ASP_DOC_CERT.equalsIgnoreCase(mMessageType)) {
                            iVTapManager.removePKIFunction(PKI_FUNC_ID_DOCSIGN);

                        }
                        JSONObject obj = getSuccessCbValues(CB_CERT_DOWNLOAD, String.valueOf(certDownloadResult));
                        sendPluginResult(mCallbackContext, PluginResult.Status.ERROR, obj);
                    }
                }
                else {
                    JSONObject obj = new JSONObject();
                    try {
                        obj.put("mMessageId", mMessageId);
                        obj.put("mMessageType", mMessageType);
                        obj.put("mNotifyMsgFlag", mNotifyMsgFlag);
                        obj.put("mNotifyMsg", mNotifyMsg);
                        obj.put("mMsgFlag", mMsgFlag);
                        obj.put("mPassType", mPassType);
                        obj.put("mNotifyMsgBytes", mNotifyMsgBytes);
                        // Encode array to Base64
                        String encodedString = android.util.Base64.encodeToString(mNotifyMsgBytes, android.util.Base64.DEFAULT);
                        obj.put("mNotifyMsgBytes", encodedString);
                        obj.put("mDataToBeSigned", mDataToBeSigned);

                    } catch (JSONException e) {
                        e.printStackTrace();
                    }

                    JSONObject outerObj = getSuccessCbValues(CB_PUSH_NOTIFICATION, obj);
                    sendPluginResult(mCallbackContext, PluginResult.Status.OK, outerObj);
                }
            }
        }
    };

    public void getInstance(final  CallbackContext callbackContext) {
        Activity activity = mPlugin.cordova.getActivity();
        try {
            iVTapManager = VTapFactory.getInstance(activity);

            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, RET_MESSAGE_SUCCESS);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void setUpVTap(final CallbackContext callbackContext, String provServer, String vtapServer, String pkiServer) {
        mCallbackContext = callbackContext;
        Activity activity = mPlugin.cordova.getActivity();

        // iVTapManager = VTapFactory.getInstance(activity);

        VGuardBroadcastReceiver mVguardRcvr = new VGuardBroadcastReceiver(activity) {
            @Override
            public void onReceive(Context context, Intent intent) {
                super.onReceive(context, intent);

                if (ACTION_SCAN_COMPLETE.equals(intent.getAction())) {
                    ArrayList<Parcelable> detectedThreats = (ArrayList<Parcelable>) intent.
                            getParcelableArrayListExtra(VGuardBroadcastReceiver.SCAN_COMPLETE_RESULT);
                    List<BasicThreatInfo> list = new ArrayList<BasicThreatInfo>();

                    for (Parcelable info : detectedThreats) {
                        BasicThreatInfo threatinfo = (BasicThreatInfo) info;
                        list.add(threatinfo);
                    }

                    JSONObject obj = getSuccessCbValues(CB_SCAN_COMPLETE, list);
                    sendPluginResult(mCallbackContext, PluginResult.Status.OK, obj);
                    Log.d(TAG, "adb_test scan completed..");
                }
            }
        };

        // Set Host Name ( provisioning server , vtap server )
        /* if (true) {
            provServer = "https://intranet.v-key.com:9575/provision";
            vtapServer = "https://intranet.v-key.com:9575/vtap";
            pkiServer = "https://intranet.v-key.com:9575";
        } else {
            provServer = "https://intranet.v-key.com:9539/provision";
            vtapServer = "https://intranet.v-key.com:9539/vtap";
            pkiServer = "https://intranet.v-key.com:9575";
        }

        // Set Host Name ( provisioning server , vtap server )
        iVTapManager.setHostName(provServer, vtapServer);

        // Set PKI Host name
        iVTapManager.setPKIHostName(pkiServer); */

        LocalBroadcastManager localBroadcastManager = LocalBroadcastManager.getInstance(activity);
        /* localBroadcastManager.registerReceiver(mVguardRcvr, new IntentFilter(VGuardBroadcastReceiver.ACTION_SCAN_COMPLETE));
        localBroadcastManager.registerReceiver(mVguardRcvr, new IntentFilter(VGuardBroadcastReceiver.ACTION_FINISH));
        localBroadcastManager.registerReceiver(mVguardRcvr, new IntentFilter(PROFILE_LOADED_ACTION));
        localBroadcastManager.registerReceiver(mVguardRcvr, new IntentFilter(VOS_READY_ACTION));
        localBroadcastManager.registerReceiver(mVguardRcvr, new IntentFilter(PROFILE_THREAT_RESPONSE)); */

        localBroadcastManager.registerReceiver(mVtapRcvr, new IntentFilter(VTAP_SETUP_ACTION));
        localBroadcastManager.registerReceiver(mPushRcvr, new IntentFilter(PUSH_NOTIFICATION_BROADCAST));
        isRegistered = true;

        iVTapManager.setupVTap();
    }

    public void setHostName(final CallbackContext callbackContext, String provServer, String vtapServer) {

        try {
            iVTapManager.setHostName(provServer, vtapServer);

            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, RET_MESSAGE_SUCCESS);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void setPKIHostName(final CallbackContext callbackContext, String pkiServer) {
        try {
            iVTapManager.setPKIHostName(pkiServer);

            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, RET_MESSAGE_SUCCESS);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void getTrustedTime(final CallbackContext callbackContext) {
        try {
            int trustedTime = iVTapManager.getTrustedTime();
            if(callbackContext != null) {
                if(trustedTime > 0 && trustedTime != ResultCode.VTAP_TRUSTED_TIME_TIMEOUT) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, trustedTime);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, trustedTime);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void checkDeviceCompatibility(final CallbackContext callbackContext) {
        try {
            int status = iVTapManager.checkDeviceCompatibility();

            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, status);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void sendDeviceInfo(final CallbackContext callbackContext, String deviceInfo, int status) {
        try {
            int result = iVTapManager.sendDeviceInfo(deviceInfo, status);
            if (callbackContext != null) {
                if (result == ResultCode.VTAP_SEND_DEVICE_INFO_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                } else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void isProvisioningDone(final CallbackContext callbackContext) {
        try {
            boolean provisioningStatus = iVTapManager.isProvisioningDone();
            sendPluginResult(callbackContext, provisioningStatus);
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }

    }

    public void getAllProvisionedTokens(final CallbackContext callbackContext) {
        try {
            String[] allProvisionedTokens = iVTapManager.getAllProvisionedTokens();
            if (callbackContext != null) {
                JSONArray arr;
                if(allProvisionedTokens != null) {
                    arr = new JSONArray(Arrays.asList(allProvisionedTokens));
                }
                else {
                    arr = new JSONArray();
                }
                sendPluginResult(callbackContext, PluginResult.Status.OK, arr);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void getTokenType(CallbackContext callbackContext, String tokenSerial) {
        try {
            VTapInterface.TokenType tokenTypeResult = iVTapManager.getTokenType(tokenSerial);
            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, tokenTypeResult.getValue());
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void getDFPHash(CallbackContext callbackContext) {
        try {
            String DFPHash = iVTapManager.getDFPHash();

            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, DFPHash);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void getLoadAckTokenFirmware(final CallbackContext callbackContext, ArrayList<String> provisioningInfo) {

        try {
            int getTokenFirmwareStatus = iVTapManager.getLoadAckTokenFirmware((provisioningInfo));
            if (callbackContext != null) {
                if (getTokenFirmwareStatus == ResultCode.VTAP_LOAD_FIRMWARE_SUCCESS || getTokenFirmwareStatus == ResultCode.VTAP_TOKEN_DOWNLOAD_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, getTokenFirmwareStatus);
                } else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, getTokenFirmwareStatus);
                }
            }

        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void loadTokenFirmware(final CallbackContext callbackContext, ArrayList<String> provisioningInfo, String tokenFilename) {

        Activity activity = mPlugin.cordova.getActivity();
        // Copy files from asset to internal storage
        String tokenFilePath = activity.getFilesDir().getPath() + File.separator + tokenFilename;
        try {
            OutputStream outputStream = new FileOutputStream(tokenFilePath);
            byte[] buffer = new byte[1024];
            int length;
            InputStream inputStream = activity.getAssets().open(tokenFilename);

            while ((length = inputStream.read(buffer)) > 0) {
                outputStream.write(buffer, 0, length);
            }

            inputStream.close();
            outputStream.flush();
            outputStream.close();

        } catch (FileNotFoundException e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        } catch (IOException e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }


        try {
            int loadTokenFirmwareStatus = iVTapManager.loadTokenFirmware(provisioningInfo.get(0), provisioningInfo.get(1), tokenFilePath);
            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, loadTokenFirmwareStatus);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void validateCheckSum(final CallbackContext callbackContext, ArrayList<String> provisioningInfo) {

        try {
            boolean checksumStatus = iVTapManager.validateCheckSum(provisioningInfo);
            sendPluginResult(callbackContext, checksumStatus);
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void getPKITokenSerial(final CallbackContext callbackContext) {
        try {
            String tokenSerial = iVTapManager.getPKITokenSerial();
            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, tokenSerial);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void getTokenSerial(final CallbackContext callbackContext) {
        try {
            String tokenSerial = iVTapManager.getTokenSerial();
            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, tokenSerial);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void getTokenFirmwareVersion(final CallbackContext callbackContext, String tokenSerial) {
        try {
            String firmwareVersion = iVTapManager.getTokenFirmwareVersion(tokenSerial);
            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, firmwareVersion);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void removeTokenFirmware(final CallbackContext callbackContext, String tokenSerial) {
        try {
            int removeTokenResult = iVTapManager.removeTokenFirmware(tokenSerial);
            if (callbackContext != null) {
                if (removeTokenResult == ResultCode.VTAP_REMOVE_TOKEN_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, removeTokenResult);
                } else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, removeTokenResult);
                }
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void loadToken(final CallbackContext callbackContext, String tokenSerial) {
        try {
            int loadTokenResult = iVTapManager.loadToken(tokenSerial);
            if (callbackContext != null) {
                if (loadTokenResult == ResultCode.VTAP_LOAD_TOKEN_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, loadTokenResult);
                } else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, loadTokenResult);
                }
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void unloadToken(final CallbackContext callbackContext, String tokenSerial) {
        try {
            int unloadTokenResult = iVTapManager.unloadToken(tokenSerial);
            if (callbackContext != null) {
                if (unloadTokenResult == ResultCode.VTAP_UNLOAD_TOKEN_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, unloadTokenResult);
                } else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, unloadTokenResult);
                }
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void createTokenPin(final CallbackContext callbackContext, String pin, String tokenSerial) {
        // Create Token Pin
        try {
            int createTokenPinStatus = iVTapManager.createTokenPIN(pin, tokenSerial);
            if (callbackContext != null) {
                if(createTokenPinStatus == ResultCode.VTAP_CREATE_PIN_SUCCESS || createTokenPinStatus == ResultCode.VTAP_PIN_ALREADY_EXIST) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, createTokenPinStatus);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, createTokenPinStatus);
                }
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void isTokenPinRemembered(final CallbackContext callbackContext, String tokenSerial) {
        try {
            boolean isTokenPinRememberedStatus = iVTapManager.isTokenPINRemembered(tokenSerial);
            if(callbackContext != null) {
                sendPluginResult(callbackContext, isTokenPinRememberedStatus);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void checkTokenPin(final CallbackContext callbackContext, String pin, boolean rememberPin, String tokenSerial) {
        try {
            int checkTokenPinStatus = iVTapManager.checkTokenPIN(pin, rememberPin, tokenSerial);
            if(callbackContext != null) {
                if(checkTokenPinStatus == ResultCode.VTAP_CHECK_PIN_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, checkTokenPinStatus);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, checkTokenPinStatus);
                }
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void checkTokenPin(final CallbackContext callbackContext, String pin, boolean rememberPin, String tokenSerial, int seedNumber) {
        try {
            int checkTokenPinStatus = iVTapManager.checkTokenPIN(pin, rememberPin, tokenSerial, seedNumber);
            if(callbackContext != null) {
                if(checkTokenPinStatus == ResultCode.VTAP_CHECK_PIN_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, checkTokenPinStatus);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, checkTokenPinStatus);
                }
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void changeTokenPin(final CallbackContext callbackContext, String oldPin, String newPin, String tokenSerial) {
        try {
            int changePinStatus = iVTapManager.changeTokenPIN(oldPin, newPin, tokenSerial);
            if(callbackContext != null) {
                if (changePinStatus == ResultCode.VTAP_CHANGE_PIN_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, changePinStatus);
                } else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, changePinStatus);
                }
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void isTokenRegistered(final CallbackContext callbackContext, String tokenSerial) {
        try {
            boolean isTokenRegisteredStatus = iVTapManager.isTokenRegistered(tokenSerial);
            if(callbackContext != null) {
                sendPluginResult(callbackContext, isTokenRegisteredStatus);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }


    public void setOtpLength(final CallbackContext callbackContext, int otpLength) {
        try {
            int setOTPLengthStatus = iVTapManager.setOtpLength(otpLength);
            if(callbackContext != null) {
                if (setOTPLengthStatus == ResultCode.VTAP_SET_OTP_LENGTH_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, setOTPLengthStatus);
                } else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, setOTPLengthStatus);
                }
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void generateTOTP(final CallbackContext callbackContext, int keyNum) {
        try {
            ArrayList<String> totpResult = iVTapManager.generateTOTP(keyNum);
            JSONArray arr = new JSONArray(totpResult);
            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, arr);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void generateCR(final CallbackContext callbackContext, String challengeValue) {
        try {
            ArrayList<String> crResult = iVTapManager.generateCR(challengeValue);
            JSONArray arr = new JSONArray(crResult);
            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, arr);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void generateTxS(final CallbackContext callbackContext, String account, String amount) {
        try {
            ArrayList<String> txsResult = iVTapManager.generateTxS(account, amount);
            JSONArray arr = new JSONArray(txsResult);
            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, arr);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void generateCsrAndSend(final CallbackContext callbackContext, int functionId, DistinguishedName distinguishedName, String pin, boolean rememberPin) {
        try {
            int generateCsrAndSendResult = iVTapManager.generateCsrAndSend(functionId, distinguishedName, pin, rememberPin);
            if (callbackContext != null) {
                if (generateCsrAndSendResult == ResultCode.VTAP_PKI_CSR_REGISTRATION_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, generateCsrAndSendResult);
                } else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, generateCsrAndSendResult);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void triggerCertPushNotification(final CallbackContext callbackContext, int functionId) {
        try {
            int result = iVTapManager.triggerCertPushNotification(functionId);
            if(callbackContext != null) {
                if (result == ResultCode.VTAP_PKI_TRIGGER_SEND_CERT_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void pkiCertDownload(final CallbackContext callbackContext, int functionId, String messageId, String messageType) {
        try {
            int result = iVTapManager.pkiCertDownload(functionId, messageId, messageType);
            if(callbackContext != null) {
                if(result == ResultCode.VTAP_PKI_CERT_DOWNLOAD_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void generateCsrAndSendSync(final CallbackContext callbackContext, int functionId, DistinguishedName distinguishedName, String pin, boolean rememberPin) {
        try {
            int result = iVTapManager.generateCsrAndSendSync(functionId, distinguishedName, pin, rememberPin);
            if(callbackContext != null) {
                if(result == ResultCode.VTAP_PKI_CSR_REGISTRATION_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void isPKIFunctionRegistered(final CallbackContext callbackContext, int functionId) {
        try {
            boolean result = iVTapManager.isPKIFunctionRegistered(functionId);
            if(callbackContext != null) {
                sendPluginResult(callbackContext, result);
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void isPKIFunctionPinRemembered(final CallbackContext callbackContext, int functionId) {
        try {
            boolean result = iVTapManager.isPKIFunctionPINRemembered(functionId);
            if(callbackContext != null) {
               sendPluginResult(callbackContext,result);
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void pkiFunctionCheckPin(final CallbackContext callbackContext, int functionId, String pin, boolean rememberPin) {
        try {
            int result = iVTapManager.pkiFunctionCheckPIN(functionId, pin, rememberPin);
            if(callbackContext != null) {
                if(result == ResultCode.VTAP_PKI_CHECK_PIN_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }
    public void pkiFunctionChangePin(CallbackContext callbackContext, int functionId, String oldPin, String newPin) {
        try {
            int result = iVTapManager.pkiFunctionChangePIN(functionId, oldPin, newPin);
            if(callbackContext != null) {
                if(result == ResultCode.VTAP_PKI_CHANGE_PIN_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void removePKIFunction(final CallbackContext callbackContext, int functionId) {
        try {
            iVTapManager.removePKIFunction(functionId);
            if(callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, RET_MESSAGE_SUCCESS);
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void pushNotificationRegister(final CallbackContext callbackContext, String userId, String deviceId, String pushToken) {
        try {
            int result = iVTapManager.pushNotificationRegister(userId, deviceId, pushToken);
            if (callbackContext != null) {
                if (result == ResultCode.VTAP_PUSH_NOTIFICATION_REGISTRATION_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                } else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void getRegisteredUserId(final CallbackContext callbackContext) {
        try {
            String userId = iVTapManager.getRegisteredUserId();
            if(callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, userId);
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void getRegisteredDeviceId(final CallbackContext callbackContext) {
        try {
            String deviceId = iVTapManager.getRegisteredDeviceId();
            if(callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, deviceId);
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void pkiFunctionAuthenticate(final CallbackContext callbackContext, String messageId, String dataToBeSigned, boolean reject){
        try {
            int result = iVTapManager.pkiFunctionAuthenticate(messageId, dataToBeSigned, reject);
            if(callbackContext != null) {
                if(result == ResultCode.VTAP_PKI_AUTHENTICATION_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void pkiFunctionDocSign(final CallbackContext callbackContext, String messageId, String dataToBeSigned, boolean reject) {
        try {
            int result = iVTapManager.pkiFunctionDocSign(messageId, dataToBeSigned, reject);
            if(callbackContext != null) {
                if(result == ResultCode.VTAP_PKI_DOC_SIGNING_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void vMessageDecrypt(final CallbackContext callbackContext, byte[] encryptedMsg) {
        try {
            String msg = iVTapManager.vMessageDecrypt(encryptedMsg);
            if(callbackContext != null) {
                if(("" + ResultCode.VTAP_PKI_VMESSAGE_DECRYPTION_FAILED).equalsIgnoreCase(msg) ||
                        ("" + ResultCode.VTAP_INVALID_INPUT).equalsIgnoreCase(msg) ||
                        ("" + ResultCode.VTAP_INVALID_API_SEQUENCE).equalsIgnoreCase(msg)) {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, msg);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, msg);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    // TODO: Change message flag to int type
    public void vMessageDownload(final CallbackContext callbackContext, String messageId, String messageType, String messageFlag) {
        try {
            String [] msgDownloadResult = iVTapManager.vMessageDownload(messageId, messageType, messageFlag);
            JSONArray arr = new JSONArray(Arrays.asList(msgDownloadResult));
            if(callbackContext != null) {
                if(msgDownloadResult[0] != null &&
                        !msgDownloadResult[0].equals(ResultCode.VTAP_PKI_MESSAGE_DOWNLOAD_FAILED) &&
                        !msgDownloadResult[0].equals(ResultCode.VTAP_INVALID_INPUT) &&
                        !msgDownloadResult[0].equals(ResultCode.VTAP_ERROR_CONNECTION_FAILED)) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, arr);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, arr);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void vMessageAck(final CallbackContext callbackContext, String messageId) {
        try {
            int result = iVTapManager.vMessageAck(messageId);
            if(callbackContext != null) {
                if(result == ResultCode.VTAP_PKI_VMESSAGE_SEND_ACK_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void setAdditionalData(final CallbackContext callbackContext, VTapInterface.AdditionDataType type, String value) {
        try {
            int result = iVTapManager.setAdditionalData(type, value);
            if(callbackContext != null) {
                if(result == ResultCode.VTAP_CREATE_ADDITIONAL_DATA_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void getAdditionalData(final CallbackContext callbackContext, VTapInterface.AdditionDataType type) {
        try {
            String additionalData = iVTapManager.getAdditionalData(type);
            if(callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, additionalData);
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void disableAdditionalData(final CallbackContext callbackContext, VTapInterface.AdditionDataType type) {
        try {
            int result = iVTapManager.disableAdditionalData(type);
            if(callbackContext != null) {
                if(result == ResultCode.VTAP_DISABLE_ADDITIONAL_DATA) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }

    }

    public void setMessageDataType(final CallbackContext callbackContext, String data) {
        try {
            int result = iVTapManager.setMessageTypeData(data);
            if(callbackContext != null) {
                if(result == ResultCode.VTAP_CREATE_ADDITIONAL_DATA_SUCCESS) {
                    sendPluginResult(callbackContext, PluginResult.Status.OK, result);
                }
                else {
                    sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
                }
            }
        }
        catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void onPause(final CallbackContext callbackContext) {
        try {
            iVTapManager.onPause();
            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, RET_MESSAGE_SUCCESS);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void onResume(final CallbackContext callbackContext) {
        try {
            iVTapManager.onResume();

            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, RET_MESSAGE_SUCCESS);

            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());

            }
        }
    }

    public void onDestroy() {
        iVTapManager.onDestroy();
    }

    public void getTroubleShootingId(final CallbackContext callbackContext) {
        try {
            String troubleshootingId = iVTapManager.getTroubleshootingId();
            if (callbackContext != null) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, troubleshootingId);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void sendTroubleShootingLogs(final CallbackContext callbackContext) {
        try {
            int result = iVTapManager.sendTroubleshootingLogs();
            if (result == ResultCode.VTAP_SEND_TROUBLESHOOTING_LOGS_SUCCESS) {
                sendPluginResult(callbackContext, PluginResult.Status.OK, result);
            } else {
                sendPluginResult(callbackContext, PluginResult.Status.ERROR, result);
            }
        } catch (Exception e) {
            if (callbackContext != null) {
                callbackContext.error(e.getMessage());
            }
        }
    }

    public void unRegisterBroadcastReceivers() {
        if(isRegistered){
            Activity activity = mPlugin.cordova.getActivity();
            LocalBroadcastManager localBroadcastManager = LocalBroadcastManager.getInstance(activity);
            localBroadcastManager.unregisterReceiver(mVtapRcvr);
            isRegistered = false;
        }
    }

    private void sendPluginResult(CallbackContext callbackContext, PluginResult.Status status, JSONObject obj) {
        PluginResult result = new PluginResult(status, obj);
        result.setKeepCallback(true);
        callbackContext.sendPluginResult(result);
    }

    private void sendPluginResult(CallbackContext callbackContext, PluginResult.Status status, JSONArray obj) {
        PluginResult result = new PluginResult(status, obj);
        result.setKeepCallback(true);
        callbackContext.sendPluginResult(result);
    }

    private void sendPluginResult(CallbackContext callbackContext, boolean b) {
        PluginResult result;
        PluginResult.Status status = PluginResult.Status.OK;

        /* if (b) {
            status = PluginResult.Status.OK;
        } else {
            status = PluginResult.Status.ERROR;
        } */

        result = new PluginResult(status, b);
        result.setKeepCallback(true);
        callbackContext.sendPluginResult(result);
    }

    private void sendPluginResult(CallbackContext callbackContext, PluginResult.Status status, String value) {
        PluginResult result;
        result = new PluginResult(status, value);
        result.setKeepCallback(true);
        callbackContext.sendPluginResult(result);
    }

    private void sendPluginResult(CallbackContext callbackContext, PluginResult.Status status, int value) {
            PluginResult result;
            result = new PluginResult(status, value);
            result.setKeepCallback(true);
            callbackContext.sendPluginResult(result);
        }

    private JSONObject getSuccessCbValues(String action, String value) {
        JSONObject jsonObject = new JSONObject();
        try {
            jsonObject.put("action", action);
            jsonObject.put("values", value);
        } catch (JSONException e) {
            Log.w(TAG, "-- getSuccessCbValues, " + e.getMessage());
        }

        return jsonObject;
    }

    private JSONObject getSuccessCbValues(String action, List<BasicThreatInfo> values) {
        JSONObject obj = new JSONObject();
        JSONArray arr = new JSONArray();

        if (values != null && values.size() > 0) {
            for (BasicThreatInfo ti : values) {
                if (ti != null) {
                    JSONObject js = new JSONObject();
                    try {
                        js.put("threatClass", ti.getThreatClass());
                        js.put("threatName", ti.getThreatName());
                        js.put("threatInfo", ti.getThreatInfo());
                        arr.put(js);
                    } catch (JSONException e) {
                    }
                }
            }
        }

        try {
            obj.put("action", action);
            obj.put("values", arr);
        } catch (JSONException e) {
            Log.w(TAG, "-- getSuccessCbValues, " + e.getMessage());
        }
        return obj;
    }

    private JSONObject getSuccessCbValues(String action, ArrayList<String> values) {
        JSONObject obj = new JSONObject();
        JSONArray arr = new JSONArray();

        if (values != null && values.size() > 0) {
            for (String ti : values) {
                if (ti != null) {
                    arr.put(ti);
                }
            }
        }

        try {
            obj.put("action", action);
            obj.put("values", arr);
        } catch (JSONException e) {
            Log.w(TAG, "-- getSuccessCbValues, " + e.getMessage());
        }
        return obj;
    }

    private JSONObject getSuccessCbValues(String action, boolean value) {
        JSONObject jsonObject = new JSONObject();
        try {
            jsonObject.put("action", action);
            jsonObject.put("values", value);
        } catch (JSONException e) {
            Log.w(TAG, "-- getSuccessCbValues, " + e.getMessage());
        }

        return jsonObject;
    }

    private JSONObject getSuccessCbValues(String action, String[] values) {
        JSONObject obj = new JSONObject();
        JSONArray arr = new JSONArray(Arrays.asList(values));

        try {
            obj.put("action", action);
            obj.put("values", arr);
        } catch (JSONException e) {
            Log.w(TAG, "-- getSuccessCbValues, " + e.getMessage());
        }
        return obj;
    }

    private JSONObject getSuccessCbValues(String action, JSONObject value) {
        JSONObject obj = new JSONObject();
        try {
            obj.put("action", action);
            obj.put("values", value);
        } catch (JSONException e) {
            Log.w(TAG, "-- getSuccessCbValues, " + e.getMessage());
        }
        return obj;
    }

    public static VTapSetupPlugin getsInstance(CordovaPlugin cordovaPlugin) {
        if (sInstance == null) {
            synchronized (VTapSetupPlugin.class) {
                if (sInstance == null) {
                    sInstance = new VTapSetupPlugin(cordovaPlugin);
                }
            }
        }
        return sInstance;
    }
}
